pub use private::Blocktree;

mod private {
    use btserde::{read_from, write_to};
    use fuse_backend_rs::{
        abi::fuse_abi::{stat64, statvfs64, Attr, CreateIn},
        api::filesystem::{
            Context, DirEntry as FuseDirEntry, Entry, FileSystem, FsOptions, OpenOptions,
        },
    };
    use log::{debug, error};
    use serde::{Deserialize, Serialize};
    use std::{
        collections::hash_map::{self, HashMap},
        ffi::CStr,
        fs::File,
        io::{self, SeekFrom, Write},
        path::{Path, PathBuf},
        sync::{
            atomic::{AtomicU64, AtomicUsize, Ordering},
            RwLock,
        },
        time::Duration,
    };

    use crate::{
        crypto::Creds, Block, BlockOpenOptions, BlockPath, BoxInIoErr, DirEntry, Directory, Epoch,
        Error, Result,
    };

    type Inode = u64;

    #[repr(u64)]
    pub enum SpecInodes {
        RootDir = 1,
        Sb = 2,
        FirstFree = 11,
    }

    #[derive(Debug, Clone, Hash, PartialEq, Eq, Serialize, Deserialize)]
    #[repr(u32)]
    /// The type of a file (regular, directory, etc).
    enum FileType {
        /// Directory.
        Dir = Self::S_IFDIR,
        /// Regular file.
        Reg = Self::S_IFREG,
    }

    impl FileType {
        const S_IFDIR: u32 = libc::S_IFDIR;
        const S_IFREG: u32 = libc::S_IFREG;

        /// Returns the underlying mode bits for this file type.
        fn value(self) -> u32 {
            self as u32
        }

        /// Attempts to convert the given mode bits into a `FileType` enum value.
        fn from_value(value: u32) -> Result<Self> {
            match value {
                FileType::S_IFDIR => Ok(FileType::Dir),
                FileType::S_IFREG => Ok(FileType::Reg),
                _ => Err(Error::custom(format!("unknown file type: 0o{value:0o}"))),
            }
        }
    }

    impl From<FileType> for u32 {
        fn from(file_type: FileType) -> Self {
            file_type.value()
        }
    }

    impl TryFrom<u32> for FileType {
        type Error = crate::Error;
        fn try_from(value: u32) -> Result<Self> {
            Self::from_value(value)
        }
    }

    impl From<SpecInodes> for Inode {
        fn from(special: SpecInodes) -> Self {
            special as Inode
        }
    }

    trait SeekFromExt {
        /// Converts the c-style `(whence, offset)` pair into a [SeekFrom] enum value.
        /// See the POSIX man page for `lseek` for more details.
        fn whence_offset(whence: u32, offset: u64) -> io::Result<SeekFrom> {
            let whence = whence as i32;
            match whence {
                libc::SEEK_SET => Ok(SeekFrom::Start(offset)),
                libc::SEEK_CUR => Ok(SeekFrom::Current(offset as i64)),
                libc::SEEK_END => Ok(SeekFrom::End(-(offset as i64))),
                _ => Err(io::Error::new(
                    io::ErrorKind::InvalidInput,
                    "`whence` was not one of `libc::{SEEK_SET, SEEK_CUR, SEEK_END}`",
                )),
            }
        }
    }

    impl SeekFromExt for SeekFrom {}

    struct InodeTableValue {
        block: Option<Box<dyn Block>>,
        /// The path to the block. This is cloned from the block's metadata when the entry is
        /// created so that it's accessible even after the block is closed.
        path: BlockPath,
        ref_count: AtomicUsize,
    }

    impl InodeTableValue {
        fn new(block: Box<dyn Block>) -> InodeTableValue {
            let path = block.meta_body().path.to_owned();
            Self {
                block: Some(block),
                path,
                ref_count: AtomicUsize::new(1),
            }
        }
    }

    type InodeTable = HashMap<Inode, InodeTableValue>;
    type InodeTableEntry<'a> = hash_map::Entry<'a, Inode, InodeTableValue>;

    /// Structure for metadata about a blocktree.
    #[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
    struct Superblock {
        /// The generation number of the cluster this part of the blocktree is stored on.
        generation: u64,
        /// The next free inode available to the cluster.
        next_inode: u64,
    }

    /// Structure for managing the part of a blocktree which is stored in the local filesystem.
    pub struct Blocktree<C> {
        /// The path to the directory in the local filesystem where this blocktree is located.
        path: PathBuf,
        /// A map from inode numbers to their reference counts.
        inodes: RwLock<InodeTable>,
        /// The next inode that will be assigned to a new block.
        next_inode: AtomicU64,
        /// The generation number of this filesystem. This is the same for every other server in
        /// the same cluster.
        generation: u64,
        /// The credentials this blocktree instance will use for all cryptographic operations.
        creds: C,
    }

    impl<C: Creds + Clone + 'static> Blocktree<C> {
        /// Creates a new empty blocktree at the given path.
        pub fn new_empty(btdir: PathBuf, generation: u64, creds: C) -> Result<Blocktree<C>> {
            let root_block_path = creds
                .writecap()
                .ok_or(Error::MissingWritecap)?
                .root_block_path();

            // Initialize the superblock.
            let mut sb_block = Self::open_block(
                &btdir,
                SpecInodes::Sb.into(),
                creds.clone(),
                root_block_path.to_owned(),
            )?;
            let sb = Superblock {
                generation,
                next_inode: SpecInodes::FirstFree.into(),
            };
            write_to(&sb, &mut sb_block)?;
            sb_block.mut_meta_body().access_secrets(|secrets| {
                secrets.inode = SpecInodes::Sb.into();
                secrets.mode = Self::default_file_mode();
                secrets.uid = Self::uid();
                secrets.gid = Self::gid();
                secrets.nlink = 1;
                Ok(())
            })?;
            sb_block.flush()?;

            // Initialize the root directory.
            let mut root_block = Self::open_block(
                &btdir,
                SpecInodes::RootDir.into(),
                creds.clone(),
                root_block_path,
            )?;
            write_to(&Directory::new(), &mut root_block)?;
            root_block.mut_meta_body().access_secrets(|secrets| {
                secrets.inode = SpecInodes::RootDir.into();
                secrets.mode = Self::default_dir_mode();
                secrets.uid = Self::uid();
                secrets.gid = Self::gid();
                secrets.nlink = 1;
                Ok(())
            })?;
            root_block.flush()?;

            Self::new(btdir, sb, sb_block, root_block, creds)
        }

        /// Opens an existing blocktree stored at the given path.
        pub fn new_existing(btdir: PathBuf, creds: C) -> Result<Blocktree<C>> {
            let root_block_path = creds
                .writecap()
                .ok_or(Error::MissingWritecap)?
                .root_block_path();
            let mut sb_block = Self::open_block(
                &btdir,
                SpecInodes::Sb.into(),
                creds.clone(),
                root_block_path.to_owned(),
            )?;
            let sb = read_from(&mut sb_block)?;
            let root_block = Self::open_block(
                &btdir,
                SpecInodes::RootDir.into(),
                creds.clone(),
                root_block_path,
            )?;
            Self::new(btdir, sb, sb_block, root_block, creds)
        }

        fn new(
            btdir: PathBuf,
            sb: Superblock,
            sb_block: Box<dyn Block>,
            root_block: Box<dyn Block>,
            creds: C,
        ) -> Result<Blocktree<C>> {
            let mut inodes = HashMap::with_capacity(1);
            inodes.insert(SpecInodes::Sb.into(), InodeTableValue::new(sb_block));
            inodes.insert(SpecInodes::RootDir.into(), InodeTableValue::new(root_block));
            Ok(Blocktree {
                path: btdir,
                inodes: RwLock::new(inodes),
                next_inode: AtomicU64::new(sb.next_inode),
                generation: sb.generation,
                creds,
            })
        }

        /// Returns the path to the file storing the given inode's data.
        fn block_path<P: AsRef<Path>>(parent: P, inode: Inode) -> PathBuf {
            let group = inode / 0xFF;
            let mut path = PathBuf::new();
            path.push(parent);
            path.push(format!("{group:02x}"));
            path.push(format!("{inode:x}.blk"));
            path
        }

        fn open_block<P: AsRef<Path>>(
            btdir: P,
            inode: Inode,
            creds: C,
            block_path: BlockPath,
        ) -> Result<Box<dyn Block>> {
            let path = Self::block_path(&btdir, inode);
            let dir = path.ancestors().nth(1).unwrap();
            if let Err(err) = std::fs::create_dir(dir) {
                match err.kind() {
                    io::ErrorKind::AlreadyExists => (),
                    _ => return Err(err.into()),
                }
            }
            let file = std::fs::OpenOptions::new()
                .read(true)
                .write(true)
                .create(true)
                .open(path)?;
            Self::open_block_file(file, creds, block_path)
        }

        fn open_block_file(file: File, creds: C, block_path: BlockPath) -> Result<Box<dyn Block>> {
            BlockOpenOptions::new()
                .with_creds(creds)
                .with_compress(false)
                .with_encrypt(true)
                .with_inner(file)
                .with_block_path(block_path)
                .open()
        }

        fn default_dir_mode() -> u32 {
            FileType::Dir.value() | 0o755
        }

        fn default_file_mode() -> u32 {
            FileType::Reg.value() | 0o644
        }

        fn uid() -> u32 {
            unsafe { libc::getuid() }
        }

        fn gid() -> u32 {
            unsafe { libc::getgid() }
        }

        /// Returns the [Err] variant containing the [io::Error] corresponding to [libc::ENOSYS].
        fn not_supported<T>() -> io::Result<T> {
            Err(io::Error::from_raw_os_error(libc::ENOSYS))
        }

        fn access_entry<T, F: FnOnce(InodeTableEntry) -> io::Result<T>>(
            &self,
            inode: Inode,
            cb: F,
        ) -> io::Result<T> {
            let mut inodes = self
                .inodes
                .write()
                .map_err(|err| Error::custom(err.to_string()))?;
            let entry = inodes.entry(inode);
            cb(entry)
        }

        fn access_value<T, F: FnOnce(&mut InodeTableValue) -> io::Result<T>>(
            &self,
            inode: Inode,
            cb: F,
        ) -> io::Result<T> {
            self.access_entry(inode, |mut entry| match &mut entry {
                InodeTableEntry::Vacant(_) => Err(Error::NotOpen(inode).into()),
                InodeTableEntry::Occupied(entry) => cb(entry.get_mut()),
            })
        }

        fn access_block<T, F: FnOnce(&mut Box<dyn Block>) -> io::Result<T>>(
            &self,
            inode: Inode,
            cb: F,
        ) -> io::Result<T> {
            self.access_value(inode, |value| {
                cb(value.block.as_mut().ok_or(Error::NotOpen(inode))?)
            })
        }

        fn access_block_path<T, F: FnOnce(&BlockPath) -> io::Result<T>>(
            &self,
            inode: Inode,
            cb: F,
        ) -> io::Result<T> {
            self.access_value(inode, |value| cb(&value.path))
        }

        /// Opens the given inode if necessary, then accesses the value it refers to.
        /// The refcount for the inode is incremented in either case (in the case the inode
        /// is opened it goes from 0 to 1).
        fn open_then_access_value<T, F: FnOnce(&mut InodeTableValue) -> io::Result<T>>(
            &self,
            inode: Inode,
            block_path: BlockPath,
            cb: F,
        ) -> io::Result<T> {
            self.access_entry(inode, |entry| match entry {
                InodeTableEntry::Vacant(entry) => {
                    let block =
                        Self::open_block(&self.path, inode, self.creds.clone(), block_path)?;
                    let mut value = InodeTableValue::new(block);
                    let result = cb(&mut value);
                    entry.insert(value);
                    result
                }
                InodeTableEntry::Occupied(mut entry) => {
                    entry.get().ref_count.fetch_add(1, Ordering::SeqCst);
                    let value = entry.get_mut();
                    if value.block.is_none() {
                        let block =
                            Self::open_block(&self.path, inode, self.creds.clone(), block_path)?;
                        value.block = Some(block);
                    }
                    cb(value)
                }
            })
        }

        fn open_then_access_block<T, F: FnOnce(&mut Box<dyn Block>) -> io::Result<T>>(
            &self,
            inode: Inode,
            block_path: BlockPath,
            cb: F,
        ) -> io::Result<T> {
            self.open_then_access_value(inode, block_path, |value| {
                cb(value.block.as_mut().ok_or(Error::NotOpen(inode))?)
            })
        }

        fn inode_open(&self, inode: Inode, block_path: BlockPath) -> io::Result<()> {
            self.open_then_access_value(inode, block_path, |_| Ok(()))
        }

        fn inode_release(&self, inode: Inode) -> io::Result<()> {
            let prev = {
                let inodes = self
                    .inodes
                    .read()
                    .map_err(|err| Error::custom(err.to_string()))?;
                let value = match inodes.get(&inode) {
                    Some(value) => value,
                    None => return Ok(()),
                };
                value.ref_count.fetch_sub(1, Ordering::SeqCst)
            };
            if 1 == prev {
                let mut inodes = self
                    .inodes
                    .write()
                    .map_err(|err| Error::custom(err.to_string()))?;
                let entry = inodes
                    .get_mut(&inode)
                    .ok_or_else(|| Error::custom("inode table entry was removed unexpectedly"))?;
                entry.block = None;
            }
            Ok(())
        }

        fn ser_write<T: Serialize>(&self, inode: Inode, value: &T) -> io::Result<()> {
            self.access_block(inode, |block| {
                block.seek(SeekFrom::Start(0))?;
                write_to(&value, block)?;
                Ok(())
            })
        }

        fn directory_read(&self, inode: Inode) -> io::Result<(Directory, Attr, BlockPath)> {
            self.access_block(inode, |block| {
                block.seek(SeekFrom::Start(0))?;
                let dir = read_from(block)?;
                let attr = block
                    .mut_meta_body()
                    .access_secrets(|secrets| Ok(secrets.attr()))?;
                let path = block.meta_body().path.to_owned();
                Ok((dir, attr, path))
            })
        }

        /// Updates the superblock file with the next valid inode value.
        fn update_sb(&self) -> io::Result<()> {
            let sb = Superblock {
                generation: self.generation,
                next_inode: self.next_inode.load(Ordering::SeqCst),
            };
            self.ser_write(SpecInodes::Sb.into(), &sb)
        }

        fn attr_timeout(&self) -> Duration {
            Duration::from_secs(5)
        }

        fn entry_timeout(&self) -> Duration {
            Duration::from_secs(5)
        }
    }

    unsafe impl<C: Sync> Sync for Blocktree<C> {}

    impl<C: Creds + Clone + 'static> FileSystem for Blocktree<C> {
        type Inode = Inode;
        type Handle = u64;

        fn init(&self, _capable: FsOptions) -> io::Result<FsOptions> {
            debug!("Blocktree::init called");
            Ok(FsOptions::empty())
        }

        fn destroy(&self) {
            debug!("Blocktree::destroy called");
        }

        fn lookup(&self, ctx: &Context, parent: Self::Inode, name: &CStr) -> io::Result<Entry> {
            debug!("Blocktree::lookup called on parent {parent}");
            let (dir, _, mut block_path) = self.directory_read(parent)?;
            let name = name.to_str().box_err()?;
            let entry = dir
                .entries
                .get(name)
                .ok_or_else(|| io::Error::from_raw_os_error(libc::ENOENT))?;
            let inode = match entry {
                DirEntry::File(entry) => entry.inode,
                DirEntry::Directory(entry) => entry.inode,
                DirEntry::Server(_) => {
                    return Err(io::Error::new(
                        io::ErrorKind::Unsupported,
                        "can't lookup server entry",
                    ))
                }
            };
            block_path.push_component(name.to_owned());
            let stat = self.open_then_access_block(inode, block_path, |block| {
                Ok(block.meta_body().secrets()?.stat())
            })?;
            Ok(Entry {
                inode,
                generation: self.generation,
                attr: stat,
                attr_flags: 0,
                attr_timeout: self.attr_timeout(),
                entry_timeout: self.entry_timeout(),
            })
        }

        fn open(
            &self,
            ctx: &Context,
            inode: Self::Inode,
            flags: u32,
            fuse_flags: u32,
        ) -> io::Result<(Option<Self::Handle>, OpenOptions)> {
            debug!("Blocktree::open called on inode {inode}");
            let block_path = self.access_block_path(inode, |p| Ok(p.to_owned()))?;
            self.inode_open(inode, block_path)?;
            Ok((None, OpenOptions::empty()))
        }

        fn release(
            &self,
            ctx: &Context,
            inode: Self::Inode,
            flags: u32,
            _handle: Self::Handle,
            flush: bool,
            flock_release: bool,
            lock_owner: Option<u64>,
        ) -> io::Result<()> {
            debug!("Blocktree::release called on inode {inode}");
            if flush {
                self.access_block(inode, |entry| entry.flush())?;
            };
            self.inode_release(inode)
        }

        fn opendir(
            &self,
            ctx: &Context,
            inode: Self::Inode,
            flags: u32,
        ) -> io::Result<(Option<Self::Handle>, OpenOptions)> {
            debug!("Blocktree::opendir called on inode {inode}");
            let block_path = self.access_block_path(inode, |p| Ok(p.to_owned()))?;
            self.inode_open(inode, block_path)?;
            Ok((None, OpenOptions::empty()))
        }

        fn releasedir(
            &self,
            ctx: &Context,
            inode: Self::Inode,
            flags: u32,
            _handle: Self::Handle,
        ) -> io::Result<()> {
            debug!("Blocktree::releasedir called for inode {inode}");
            self.inode_release(inode)
        }

        fn create(
            &self,
            ctx: &Context,
            parent: Self::Inode,
            name: &CStr,
            args: CreateIn,
        ) -> std::io::Result<(Entry, Option<Self::Handle>, OpenOptions)> {
            debug!("Blocktree::create called on parent {parent}");

            let name = name
                .to_str()
                .map_err(|err| io::Error::new(io::ErrorKind::InvalidInput, err))?
                .to_owned();

            // Reserve a free inode.
            // TODO: Obviously this strategy won't work when there are multiple servers in this
            // generation.
            let inode = self.next_inode.fetch_add(1, Ordering::SeqCst);
            self.update_sb()?;

            // Create the new file block.
            let mut block_path =
                self.access_block(parent, |block| Ok(block.meta_body().path.to_owned()))?;
            block_path.push_component(name.to_owned());
            self.inode_open(inode, block_path)?;

            // Add a directory entry to the parent for the new inode.
            self.access_block(parent, |dir_block| {
                dir_block.seek(SeekFrom::Start(0))?;
                let mut dir: Directory = read_from(dir_block).box_err()?;
                dir.add_file(name, inode)?;
                dir_block.seek(SeekFrom::Start(0))?;
                write_to(&dir, dir_block).box_err()?;
                dir_block.flush()?;
                Ok(())
            })?;

            let attr = self.access_block(inode, |block| {
                Ok(block.mut_meta_body().access_secrets(|secrets| {
                    secrets.inode = inode;
                    secrets.mode = Self::default_file_mode();
                    secrets.uid = Self::uid();
                    secrets.gid = Self::gid();
                    let now = Epoch::now();
                    secrets.atime = now;
                    secrets.ctime = now;
                    secrets.mtime = now;
                    secrets.nlink = 1;
                    Ok(secrets.attr())
                })?)
            })?;
            let entry = Entry {
                inode,
                generation: self.generation,
                attr: attr.into(),
                attr_flags: 0,
                attr_timeout: self.attr_timeout(),
                entry_timeout: self.entry_timeout(),
            };
            // TODO: What is OpenOptions used for?
            let options = OpenOptions::empty();
            Ok((entry, None, options))
        }

        fn write(
            &self,
            ctx: &Context,
            inode: Self::Inode,
            _handle: Self::Handle,
            r: &mut dyn fuse_backend_rs::api::filesystem::ZeroCopyReader,
            size: u32,
            offset: u64,
            lock_owner: Option<u64>,
            delayed_write: bool,
            flags: u32,
            fuse_flags: u32,
        ) -> io::Result<usize> {
            debug!("Blocktree::write called called on inode {inode}");
            let mut size: usize = size.try_into().box_err()?;
            self.access_block(inode, |block| {
                let mut buf = [0u8; crate::SECTOR_SZ_DEFAULT];
                let mut written = 0;
                while size > 0 {
                    let read = match r.read(&mut buf) {
                        Ok(size) => size,
                        Err(err) => {
                            if written > 0 {
                                error!("error while reading: {err}");
                                return Ok(written);
                            } else {
                                return Err(err);
                            }
                        }
                    };
                    if 0 == read {
                        break;
                    }
                    let filled = &buf[..read];
                    size -= read;
                    if let Err(err) = block.write_all(filled) {
                        if written > 0 {
                            error!("error while writing: {err}");
                            return Ok(written);
                        } else {
                            return Err(err);
                        }
                    }
                    written += filled.len();
                }
                Ok(written)
            })
        }

        fn flush(
            &self,
            ctx: &Context,
            inode: Self::Inode,
            handle: Self::Handle,
            lock_owner: u64,
        ) -> io::Result<()> {
            debug!("Blocktree::flush called for inode {inode}");
            self.access_block(inode, |block| block.flush())
        }

        fn read(
            &self,
            ctx: &Context,
            inode: Self::Inode,
            _handle: Self::Handle,
            w: &mut dyn fuse_backend_rs::api::filesystem::ZeroCopyWriter,
            size: u32,
            offset: u64,
            lock_owner: Option<u64>,
            flags: u32,
        ) -> io::Result<usize> {
            debug!("Blocktree::read called on inode {inode}");
            let mut size: usize = size.try_into().box_err()?;
            self.access_block(inode, |block| {
                let mut buf = [0u8; crate::SECTOR_SZ_DEFAULT];
                let mut read = 0;
                while size > 0 {
                    let just_read = match block.read(&mut buf) {
                        Ok(just_read) => just_read,
                        Err(err) => {
                            if read > 0 {
                                error!("error while reading from block: {err}");
                                return Ok(read);
                            } else {
                                return Err(err);
                            }
                        }
                    };
                    if 0 == just_read {
                        break;
                    }
                    read += just_read;
                    let filled = &buf[..just_read];
                    if let Err(err) = w.write_all(filled) {
                        if read > 0 {
                            error!("error while writing: {err}");
                            return Ok(read);
                        } else {
                            return Err(err);
                        }
                    }
                    size -= filled.len();
                }
                Ok(read)
            })
        }

        fn readdir(
            &self,
            ctx: &Context,
            inode: Self::Inode,
            _handle: Self::Handle,
            size: u32,
            offset: u64,
            add_entry: &mut dyn FnMut(
                fuse_backend_rs::api::filesystem::DirEntry,
            ) -> io::Result<usize>,
        ) -> io::Result<()> {
            debug!("Blocktree::readdir called on inode {inode}");
            let dir = self.access_block(inode, |block| {
                block.seek(SeekFrom::Start(0))?;
                Ok(read_from::<Directory, _>(block)?)
            })?;
            let mut index = 0u64;
            for (name, entry) in dir.entries() {
                index += 1;
                // TODO: Using a simple index will fail when the directory is being modified
                // concurrently.
                if index <= offset {
                    continue;
                }
                let inode = match entry.inode() {
                    Some(inode) => inode,
                    None => continue,
                };
                let dir_entry = FuseDirEntry {
                    ino: inode,
                    offset: index,
                    type_: entry.kind() as u32,
                    name: name.as_bytes(),
                };
                add_entry(dir_entry)?;
            }
            Ok(())
        }

        fn getattr(
            &self,
            ctx: &Context,
            inode: Self::Inode,
            _handle: Option<Self::Handle>,
        ) -> io::Result<(stat64, Duration)> {
            debug!("Blocktree::getattr called for inode {inode}");
            let mut stat = self.access_block(inode, |block| {
                Ok(block
                    .mut_meta_body()
                    .access_secrets(|secrets| Ok(secrets.stat()))?)
            })?;
            stat.st_ino = inode;
            Ok((stat, self.attr_timeout()))
        }

        fn forget(&self, ctx: &Context, inode: Self::Inode, count: u64) {
            debug!("Blocktree::forget called for inode {inode}");
        }

        fn lseek(
            &self,
            ctx: &Context,
            inode: Self::Inode,
            _handle: Self::Handle,
            offset: u64,
            whence: u32,
        ) -> io::Result<u64> {
            debug!("Blocktree::lseek called for inode {inode}");
            let seek_from = SeekFrom::whence_offset(whence, offset)?;
            self.access_block(inode, |block| block.seek(seek_from))
        }

        //////////////////////////////////
        // METHODS WHICH ARE NOT SUPPORTED
        //////////////////////////////////

        fn getxattr(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _name: &CStr,
            _size: u32,
        ) -> io::Result<fuse_backend_rs::api::filesystem::GetxattrReply> {
            debug!("Blocktree::getxattr called for inode {inode}");
            Self::not_supported()
        }

        fn ioctl(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _handle: Self::Handle,
            _flags: u32,
            _cmd: u32,
            _data: fuse_backend_rs::api::filesystem::IoctlData,
            _out_size: u32,
        ) -> io::Result<fuse_backend_rs::api::filesystem::IoctlData> {
            debug!("Blocktree::ioctl called for inode {inode}");
            Self::not_supported()
        }

        fn access(&self, _ctx: &Context, inode: Self::Inode, _mask: u32) -> io::Result<()> {
            debug!("Blocktree::access called for inode {inode}");
            Self::not_supported()
        }

        fn batch_forget(&self, _ctx: &Context, _requests: Vec<(Self::Inode, u64)>) {
            debug!("Blocktree::batch_forget called");
            Self::not_supported().unwrap()
        }

        fn bmap(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _block: u64,
            _blocksize: u32,
        ) -> io::Result<u64> {
            debug!("Blocktree::bmap called for inode {inode}");
            Self::not_supported()
        }

        fn fallocate(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _handle: Self::Handle,
            _mode: u32,
            _offset: u64,
            _length: u64,
        ) -> io::Result<()> {
            debug!("Blocktree::fallocate called for inode {inode}");
            Self::not_supported()
        }

        fn fsync(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _datasync: bool,
            _handle: Self::Handle,
        ) -> io::Result<()> {
            debug!("Blocktree::fsync called for inode {inode}");
            Self::not_supported()
        }

        fn fsyncdir(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _datasync: bool,
            _handle: Self::Handle,
        ) -> io::Result<()> {
            debug!("Blocktree::fsyncdir called for inode {inode}");
            Self::not_supported()
        }

        fn getlk(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _handle: Self::Handle,
            _owner: u64,
            _lock: fuse_backend_rs::api::filesystem::FileLock,
            _flags: u32,
        ) -> io::Result<fuse_backend_rs::api::filesystem::FileLock> {
            debug!("Blocktree::getlk called for inode {inode}");
            Self::not_supported()
        }

        fn link(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _newparent: Self::Inode,
            _newname: &CStr,
        ) -> io::Result<Entry> {
            debug!("Blocktree::link called for inode {inode}");
            Self::not_supported()
        }

        fn listxattr(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _size: u32,
        ) -> io::Result<fuse_backend_rs::api::filesystem::ListxattrReply> {
            debug!("Blocktree::listxattr called for inode {inode}");
            Self::not_supported()
        }

        fn mkdir(
            &self,
            _ctx: &Context,
            _parent: Self::Inode,
            _name: &CStr,
            _mode: u32,
            _umask: u32,
        ) -> io::Result<Entry> {
            debug!("Blocktree::mkdir called");
            Self::not_supported()
        }

        fn mknod(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _name: &CStr,
            _mode: u32,
            _rdev: u32,
            _umask: u32,
        ) -> io::Result<Entry> {
            debug!("Blocktree::mknod called for inode {inode}");
            Self::not_supported()
        }

        fn notify_reply(&self) -> io::Result<()> {
            debug!("Blocktree::notify_reply called");
            Self::not_supported()
        }

        fn poll(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _handle: Self::Handle,
            _khandle: Self::Handle,
            _flags: u32,
            _events: u32,
        ) -> io::Result<u32> {
            debug!("Blocktree::poll called for inode {inode}");
            Self::not_supported()
        }

        fn readdirplus(
            &self,
            _ctx: &Context,
            _inode: Self::Inode,
            _handle: Self::Handle,
            _size: u32,
            _offset: u64,
            _add_entry: &mut dyn FnMut(
                fuse_backend_rs::api::filesystem::DirEntry,
                Entry,
            ) -> io::Result<usize>,
        ) -> io::Result<()> {
            debug!("Blocktree::readdirplus called");
            Self::not_supported()
        }

        fn readlink(&self, _ctx: &Context, inode: Self::Inode) -> io::Result<Vec<u8>> {
            debug!("Blocktree::readlink called for inode {inode}");
            Self::not_supported()
        }

        fn removexattr(&self, _ctx: &Context, inode: Self::Inode, _name: &CStr) -> io::Result<()> {
            debug!("Blocktree::removexattr called for inode {inode}");
            Self::not_supported()
        }

        fn rename(
            &self,
            _ctx: &Context,
            _olddir: Self::Inode,
            _oldname: &CStr,
            _newdir: Self::Inode,
            _newname: &CStr,
            _flags: u32,
        ) -> io::Result<()> {
            debug!("Blocktree::rename called");
            Self::not_supported()
        }

        fn rmdir(&self, _ctx: &Context, parent: Self::Inode, _name: &CStr) -> io::Result<()> {
            debug!("Blocktree::rmdir called on parent {parent}");
            Self::not_supported()
        }

        fn setattr(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _attr: stat64,
            _handle: Option<Self::Handle>,
            _valid: fuse_backend_rs::api::filesystem::SetattrValid,
        ) -> io::Result<(stat64, Duration)> {
            debug!("Blocktree::setattr called for inode {inode}");
            Self::not_supported()
        }

        fn setlk(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _handle: Self::Handle,
            _owner: u64,
            _lock: fuse_backend_rs::api::filesystem::FileLock,
            _flags: u32,
        ) -> io::Result<()> {
            debug!("Blocktree::setlk called for inode {inode}");
            Self::not_supported()
        }

        fn setlkw(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _handle: Self::Handle,
            _owner: u64,
            _lock: fuse_backend_rs::api::filesystem::FileLock,
            _flags: u32,
        ) -> io::Result<()> {
            debug!("Blocktree::setlkw called for inode {inode}");
            Self::not_supported()
        }

        fn setxattr(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _name: &CStr,
            _value: &[u8],
            _flags: u32,
        ) -> io::Result<()> {
            debug!("Blocktree::setxattr called for inode {inode}");
            Self::not_supported()
        }

        fn statfs(&self, _ctx: &Context, inode: Self::Inode) -> io::Result<statvfs64> {
            debug!("Blocktree::statfs called for inode {inode}");
            Self::not_supported()
        }

        fn symlink(
            &self,
            _ctx: &Context,
            _linkname: &CStr,
            _parent: Self::Inode,
            _name: &CStr,
        ) -> io::Result<Entry> {
            debug!("Blocktree::symlink called");
            Self::not_supported()
        }

        fn unlink(&self, _ctx: &Context, parent: Self::Inode, _name: &CStr) -> io::Result<()> {
            debug!("Blocktree::unlink called on parent {parent}");
            Self::not_supported()
        }
    }
}

#[cfg(test)]
mod tests {
    use std::ffi::CString;

    use fuse_backend_rs::api::filesystem::{FileSystem, FsOptions};
    use tempdir::TempDir;

    use test_helpers::*;

    use crate::{crypto::ConcreteCreds, test_helpers, Decompose};

    use super::{private::SpecInodes, *};

    struct BtTestCase {
        dir: TempDir,
        bt: Blocktree<ConcreteCreds>,
    }

    impl BtTestCase {
        fn new_empty() -> BtTestCase {
            let dir = TempDir::new("fuse").expect("failed to create temp dir");
            let bt = Blocktree::new_empty(dir.path().to_owned(), 0, Self::creds())
                .expect("failed to create empty blocktree");
            bt.init(FsOptions::empty()).expect("init failed");
            BtTestCase { dir, bt }
        }

        fn new_existing(dir: TempDir) -> BtTestCase {
            let bt = Blocktree::new_existing(dir.path().to_owned(), Self::creds())
                .expect("failed to create blocktree from existing directory");
            bt.init(FsOptions::empty()).expect("init failed");
            BtTestCase { dir, bt }
        }

        fn creds() -> ConcreteCreds {
            test_helpers::NODE_CREDS.clone()
        }
    }

    /// Tests that a new file can be created, written to and the written data can be read from it.
    #[test]
    fn create_open_write_read() {
        let case = BtTestCase::new_empty();
        let bt = &case.bt;

        let name = CString::new("README.md").unwrap();

        let (entry, ..) = bt
            .create(
                &Default::default(),
                SpecInodes::RootDir.into(),
                name.as_c_str(),
                Default::default(),
            )
            .expect("failed to create file");

        bt.open(&Default::default(), entry.inode, 0, 0)
            .expect("failed to open file");
        const LEN: usize = 32;
        let mut expected = BtCursor::new([1u8; LEN]);

        let written = bt
            .write(
                &Default::default(),
                entry.inode,
                Default::default(),
                &mut expected,
                LEN as u32,
                0,
                None,
                false,
                0,
                0,
            )
            .expect("write failed");
        assert_eq!(LEN, written);
        let mut actual = BtCursor::new([0u8; LEN]);
        bt.lseek(&Default::default(), entry.inode, Default::default(), 0, 0)
            .expect("lseek failed");

        let read = bt
            .read(
                &Default::default(),
                entry.inode,
                Default::default(),
                &mut actual,
                LEN as u32,
                0,
                None,
                0,
            )
            .expect("failed to read");
        assert_eq!(LEN, read);

        assert_eq!(expected, actual)
    }

    #[test]
    fn lookup() {
        let case = BtTestCase::new_empty();
        let bt = &case.bt;

        let name = CString::new("README.md").unwrap();
        let (expected, ..) = bt
            .create(
                &Default::default(),
                SpecInodes::RootDir.into(),
                name.as_c_str(),
                Default::default(),
            )
            .expect("failed to create file");

        let actual = bt
            .lookup(&Default::default(), SpecInodes::RootDir.into(), &name)
            .expect("lookup failed");

        assert_eq!(expected.generation, actual.generation);
        assert_eq!(expected.inode, actual.inode);
    }

    /// Tests that data written by one instance of [Blocktree] can be read by a subsequent
    /// instance.
    #[test]
    fn new_existing() {
        const EXPECTED: &[u8] = b"cool as cucumbers";
        let name = CString::new("RESIGNATION.docx").unwrap();

        let case = BtTestCase::new_empty();
        let bt = &case.bt;

        {
            let (entry, ..) = bt
                .create(
                    &Default::default(),
                    SpecInodes::RootDir.into(),
                    name.as_c_str(),
                    Default::default(),
                )
                .expect("failed to create file");

            bt.open(&Default::default(), entry.inode, 0, 0)
                .expect("open failed");

            let mut vec = Vec::with_capacity(EXPECTED.len());
            vec.extend_from_slice(EXPECTED);
            let mut cursor = BtCursor::new(vec);
            let written = bt
                .write(
                    &Default::default(),
                    entry.inode,
                    Default::default(),
                    &mut cursor,
                    EXPECTED.len() as u32,
                    0,
                    None,
                    false,
                    0,
                    0,
                )
                .expect("write failed");
            assert_eq!(EXPECTED.len(), written);

            bt.flush(&Default::default(), entry.inode, Default::default(), 0)
                .expect("flush failed");
        }

        let case = BtTestCase::new_existing(case.dir);
        let bt = &case.bt;

        let entry = bt
            .lookup(&Default::default(), SpecInodes::RootDir.into(), &name)
            .expect("lookup failed");

        bt.open(&Default::default(), entry.inode, 0, 0)
            .expect("open failed");

        let mut actual = BtCursor::new([0u8; EXPECTED.len()]);
        let _ = bt
            .read(
                &Default::default(),
                entry.inode,
                Default::default(),
                &mut actual,
                EXPECTED.len() as u32,
                0,
                None,
                0,
            )
            .expect("read failed");

        assert_eq!(EXPECTED, actual.into_inner().as_slice())
    }
}
